@@ -20,6 +20,9 @@
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
 #endif
+
+#define PY_SIZE_T_CLEAN
+
 #include "pyorbit-private.h"
 
 size_t  ORBit_gather_alloc_info (CORBA_TypeCode tc);
@@ -234,11 +237,6 @@ marshal_value(CORBA_TypeCode tc, gconstpointer *val, PyObject *value)
 	    getval(val, CORBA_string) = CORBA_string_dup(PyString_AsString(value));
 	    advanceptr(val, sizeof(CORBA_char *));
 	    ret = TRUE;
-	} else if (value == Py_None) {
-	    alignval(val, ORBIT_ALIGNOF_CORBA_POINTER);
-	    getval(val, CORBA_string) = NULL;
-	    advanceptr(val, sizeof(CORBA_char *));
-	    ret = TRUE;
 	}
 	break;
     case CORBA_tk_sequence: 
@@ -254,20 +252,36 @@ marshal_value(CORBA_TypeCode tc, gconstpointer *val, PyObject *value)
 
 	    alignval(val, ORBIT_ALIGNOF_CORBA_SEQ);
 	    sval = (CORBA_sequence_CORBA_octet *)*val;
-	    sval->_release = TRUE;
-	    sval->_length = PySequence_Length(value);
-	    sval->_buffer = ORBit_alloc_tcval(tc->subtypes[0], sval->_length);
-	    seqval = sval->_buffer;
-	    for (i = 0; i < sval->_length; i++) {
-		PyObject *item = PySequence_GetItem(value, i);
-		gboolean itemret;
-
-		if (!item) break;
-		itemret = marshal_value(tc->subtypes[0], &seqval, item);
-		Py_DECREF(item);
-		if (!itemret) break;
+	    if (CORBA_TypeCode_equal(tc->subtypes[0], TC_CORBA_octet, NULL)
+		&& PyString_Check(value))
+	    {
+		Py_ssize_t length;
+		char *buffer;
+		if (PyString_AsStringAndSize(value, &buffer, &length) == -1)
+		    ret = FALSE;
+		else {
+		    sval->_release = CORBA_TRUE;
+		    sval->_buffer  = ORBit_alloc_tcval(TC_CORBA_octet, length);
+		    memcpy(sval->_buffer, buffer, length);
+		    sval->_length  = length;
+		    ret = TRUE;
+		}
+	    } else {
+		sval->_release = TRUE;
+		sval->_length = PySequence_Length(value);
+		sval->_buffer = ORBit_alloc_tcval(tc->subtypes[0], sval->_length);
+		seqval = sval->_buffer;
+		for (i = 0; i < sval->_length; i++) {
+		    PyObject *item = PySequence_GetItem(value, i);
+		    gboolean itemret;
+		    
+		    if (!item) break;
+		    itemret = marshal_value(tc->subtypes[0], &seqval, item);
+		    Py_DECREF(item);
+		    if (!itemret) break;
+		}
+		if (i == sval->_length) ret = TRUE; /* no error */
 	    }
-	    if (i == sval->_length) ret = TRUE; /* no error */
 	    advanceptr(val, sizeof(CORBA_sequence_CORBA_octet));
 	}
 	break;
@@ -336,11 +350,6 @@ marshal_value(CORBA_TypeCode tc, gconstpointer *val, PyObject *value)
 	    }
 	    advanceptr(val, sizeof(CORBA_wchar *));
 	    ret = TRUE;
-	} else if (value == Py_None) {
-	    alignval(val, ORBIT_ALIGNOF_CORBA_POINTER);
-	    getval(val, CORBA_wstring) = NULL;
-	    advanceptr(val, sizeof(CORBA_wchar *));
-	    ret = TRUE;
 	}
 	break;
     default:
@@ -457,6 +466,8 @@ demarshal_value(CORBA_TypeCode tc, gconstpointer *val)
 		stub = pyorbit_exception;
 	}
 	instance = PyObject_CallFunction(stub, "()");
+	if (!instance)
+	    break;
 	if (stub == pyorbit_exception) {
 	    PyObject *pytc = pycorba_typecode_new(tc);
 
@@ -495,6 +506,8 @@ demarshal_value(CORBA_TypeCode tc, gconstpointer *val)
 		stub = PyExc_Exception;
 	}
 	instance = PyObject_CallFunction(stub, "()");
+	if (!instance)
+	    break;
 
 	discrim = demarshal_value(tc->discriminator, val);
 	if (!discrim) {
@@ -556,13 +569,14 @@ demarshal_value(CORBA_TypeCode tc, gconstpointer *val)
 	switch (tc->subtypes[0]->kind) {
 	case CORBA_tk_char:
 	case CORBA_tk_octet:
-	    ret = PyString_FromStringAndSize(sval->_buffer, sval->_length);
+	    ret = PyString_FromStringAndSize((char *) sval->_buffer, sval->_length);
 	    break;
 	default: {
-	    PyObject *tuple;
-	    gint i, align;
+	    PyObject *list;
+	    Py_ssize_t i;
+            int align;
 
-	    tuple = PyTuple_New(sval->_length);
+	    list = PyList_New(sval->_length);
 	    align = tc->subtypes[0]->c_align;
 	    seqval = sval->_buffer;
 	    for (i = 0; i < sval->_length; i++) {
@@ -571,12 +585,12 @@ demarshal_value(CORBA_TypeCode tc, gconstpointer *val)
 		alignval(&seqval, align);
 		item = demarshal_value(tc->subtypes[0], &seqval);
 		if (!item) {
-		    Py_DECREF(tuple);
+		    Py_DECREF(list);
 		    break;
 		}
-		PyTuple_SetItem(tuple, i, item);
+		PyList_SetItem(list, i, item);
 	    }
-	    if (i == sval->_length) ret = tuple;
+	    if (i == sval->_length) ret = list;
 	}
 	}
 	break;
@@ -589,10 +603,10 @@ demarshal_value(CORBA_TypeCode tc, gconstpointer *val)
 	    advanceptr(val, tc->length);
 	    break;
 	default: {
-	    PyObject *tuple;
+	    PyObject *list;
 	    gint i, align;
 
-	    tuple = PyTuple_New(tc->length);
+	    list = PyList_New(tc->length);
 	    align = tc->subtypes[0]->c_align;
 	    for (i = 0; i < tc->length; i++) {
 		PyObject *item;
@@ -600,12 +614,12 @@ demarshal_value(CORBA_TypeCode tc, gconstpointer *val)
 		alignval(val, align);
 		item = demarshal_value(tc->subtypes[0], val);
 		if (!item) {
-		    Py_DECREF(tuple);
+		    Py_DECREF(list);
 		    break;
 		}
-		PyTuple_SetItem(tuple, i, item);
+		PyList_SetItem(list, i, item);
 	    }
-	    if (i == tc->length) ret = tuple;
+	    if (i == tc->length) ret = list;
 	}
 	}
 	break;
