@@ -20,6 +20,9 @@
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
 #endif
+
+#define PY_SIZE_T_CLEAN
+
 #include "pyorbit-private.h"
 
 typedef struct {
@@ -49,52 +52,41 @@ pycorba_method_repr(PyCORBA_Method *self)
 			       self->imethod->name);
 }
 
-/* XXXX handle keyword arguments? */
-static PyObject *
-pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
+static inline gboolean
+pycorba_call_marshal_args(ORBit_IMethod *imethod, PyObject *args,
+                          CORBA_TypeCode *pret_tc, gpointer *pret,
+                          gpointer **pretptr, gpointer **pargv,
+                          gpointer **pargvptr, int *pnum_args, int *p_n_rets)
 {
-    ORBit_IMethod *imethod = self->imethod;
-    CORBA_Object objref;
-    CORBA_Environment ev;
-    CORBA_TypeCode ret_tc, *arg_tc;
-    gpointer ret = NULL, retptr, *argv = NULL, *argvptr = NULL;
-    PyObject *obj, *item, *pyret = NULL;
-    gint n_args, n_rets, argpos, num_args, i, retpos;
+    CORBA_TypeCode ret_tc;
+    gpointer ret = NULL, *retptr = NULL, *argv = NULL, *argvptr = NULL;
+    PyObject *item;
+    gint n_args, n_rets, argpos, num_args, i;
     gboolean has_ret;
 
     /* unaliased argument types ... */
-    ret_tc = self->imethod->ret;
+    ret_tc = imethod->ret;
     while (ret_tc && ret_tc->kind == CORBA_tk_alias)
 	ret_tc = ret_tc->subtypes[0];
     has_ret = ret_tc != CORBA_OBJECT_NIL && ret_tc->kind != CORBA_tk_void;
-    arg_tc = g_new(CORBA_TypeCode, imethod->arguments._length);
     /* calculate number of in or inout arguments */
     n_args = 0;
     n_rets = 0;
     for (i = 0; i < imethod->arguments._length; i++) {
-	if ((self->imethod->arguments._buffer[i].flags &
+	if ((imethod->arguments._buffer[i].flags &
 	     (ORBit_I_ARG_IN | ORBit_I_ARG_INOUT)) != 0)
 	    n_args++;
-	if ((self->imethod->arguments._buffer[i].flags &
+	if ((imethod->arguments._buffer[i].flags &
 	     (ORBit_I_ARG_OUT | ORBit_I_ARG_INOUT)) != 0)
 	    n_rets++;
-
-	arg_tc[i] = self->imethod->arguments._buffer[i].tc;
-	while (arg_tc[i]->kind == CORBA_tk_alias)
-	    arg_tc[i] = arg_tc[i]->subtypes[0];
     }
     if (PyTuple_Size(args) != n_args + 1) {
-	PyErr_SetString(PyExc_TypeError, "wrong number of arguments");
-	return NULL;
+	PyErr_Format(PyExc_TypeError, "wrong number of arguments: expected %i, got %i",
+                     n_args + 1, (int) PyTuple_Size(args));
+        PyObject_Print(args, stderr, 0);
+	return FALSE;
     }
 
-    obj = PyTuple_GetItem(args, 0);
-    if (!PyObject_TypeCheck(obj, (PyTypeObject *)self->meth_class)) {
-	PyErr_SetString(PyExc_TypeError, "wrong object type as first arg");
-	return NULL;
-    }
-    objref = ((PyCORBA_Object *)obj)->objref;
-
     num_args = imethod->arguments._length;
     
     /* set up return value argument, depending on type */
@@ -103,15 +95,15 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
 	case CORBA_tk_any:
 	case CORBA_tk_sequence:
 	case CORBA_tk_array:
-	    retptr = NULL;
-	    ret = &retptr;
+            retptr = g_new0(gpointer, 1);
+	    ret = retptr;
 	    break;
 	case CORBA_tk_struct:
 	case CORBA_tk_union:
 	    /* handle non-fixed size structs, arrays and unions like any/seq */
 	    if ((imethod->flags & ORBit_I_COMMON_FIXED_SIZE) == 0) {
-		retptr = NULL;
-		ret = &retptr;
+                retptr = g_new0(gpointer, 1);
+		ret = retptr;
 		break;
 	    }
 	    /* else fall through */
@@ -121,8 +113,8 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
     argv = g_new0(gpointer, num_args);
     argvptr = g_new0(gpointer, num_args);
     for (i = 0; i < num_args; i++) {
-	gint flags = self->imethod->arguments._buffer[i].flags;
-	CORBA_TypeCode tc = self->imethod->arguments._buffer[i].tc;
+	gint flags = imethod->arguments._buffer[i].flags;
+	CORBA_TypeCode tc = imethod->arguments._buffer[i].tc;
 
 	if ((flags & (ORBit_I_ARG_IN | ORBit_I_ARG_INOUT)) != 0) {
 	    argv[i] = ORBit_small_alloc(tc);
@@ -141,7 +133,7 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
 
     argpos = 1;
     for (i = 0; i < num_args; i++) {
-	gint flags = self->imethod->arguments._buffer[i].flags;
+	gint flags = imethod->arguments._buffer[i].flags;
 
 	if ((flags & (ORBit_I_ARG_IN | ORBit_I_ARG_INOUT)) != 0) {
 	    CORBA_any any = { NULL, NULL, FALSE };
@@ -153,17 +145,27 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
 		PyErr_Format(PyExc_TypeError, "could not marshal arg '%s'",
 			     imethod->arguments._buffer[i].name ?
 			     imethod->arguments._buffer[i].name : "<unknown>");
-		goto cleanup;
+		return FALSE;
 	    }
 	}
     }
-    CORBA_exception_init(&ev);
-    ORBit_small_invoke_stub(objref, imethod, ret, argv,
-			    CORBA_OBJECT_NIL, &ev);
+    *pret = ret;
+    *pargv = argv;
+    *pargvptr = argvptr;
+    *pret_tc = ret_tc;
+    *pnum_args = num_args;
+    *pretptr = retptr;
+    *p_n_rets = n_rets;
+    return TRUE;
+}
 
-    if (pyorbit_check_ex(&ev))
-	goto cleanup;
-    CORBA_exception_free(&ev);
+static inline PyObject *
+pycorba_call_demarshal_retval(ORBit_IMethod *imethod, int num_args, CORBA_TypeCode ret_tc,
+                              gpointer ret, gpointer *argv, gpointer *argvptr, int n_rets)
+{
+    gint i, retpos;
+    PyObject *pyret = NULL, *item;
+    gboolean has_ret = (ret_tc != CORBA_OBJECT_NIL && ret_tc->kind != CORBA_tk_void);
 
     pyret = PyTuple_New(n_rets + (has_ret ? 1 : 0));
     retpos = 0;
@@ -194,12 +196,12 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
 	    pyret = NULL;
 	    PyErr_SetString(PyExc_TypeError,
 			    "could not demarshal return value");
-	    goto cleanup;
+	    return NULL;
 	}
 	PyTuple_SetItem(pyret, retpos++, item);
     }
     for (i = 0; i < num_args; i++) {
-	gint flags = self->imethod->arguments._buffer[i].flags;
+	gint flags = imethod->arguments._buffer[i].flags;
 
 	if ((flags & ORBit_I_ARG_OUT) != 0) {
 	    CORBA_any any = { NULL, NULL, FALSE };
@@ -209,12 +211,11 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
 	    item = pyorbit_demarshal_any(&any);
 	    if (!item) {
 		Py_DECREF(pyret);
-		pyret = NULL;
 		PyErr_Format(PyExc_TypeError,
 			     "could not demarshal return value '%s'",
 			     imethod->arguments._buffer[i].name ?
 			     imethod->arguments._buffer[i].name : "<unknown>");
-		goto cleanup;
+		return NULL;
 	    }
 	    PyTuple_SetItem(pyret, retpos++, item);
 	} else if ((flags & (ORBit_I_ARG_INOUT)) != 0) {
@@ -225,12 +226,11 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
 	    item = pyorbit_demarshal_any(&any);
 	    if (!item) {
 		Py_DECREF(pyret);
-		pyret = NULL;
 		PyErr_Format(PyExc_TypeError,
 			     "could not demarshal return value '%s'",
 			     imethod->arguments._buffer[i].name ?
 			     imethod->arguments._buffer[i].name : "<unknown>");
-		goto cleanup;
+		return NULL;
 	    }
 	    PyTuple_SetItem(pyret, retpos++, item);
 	}
@@ -251,20 +251,29 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
     default:
 	break;
     }
+    return pyret;
+}
 
- cleanup:
+static inline void
+pycorba_call_cleanup(ORBit_IMethod *imethod, int num_args, gpointer ret,
+                     gpointer *argv, gpointer *argvptr, gpointer *retptr,
+                     CORBA_TypeCode ret_tc)
+{
+    int i;
     if (ret) {
 	switch (ret_tc->kind) {
 	case CORBA_tk_any:
 	case CORBA_tk_sequence:
 	case CORBA_tk_array:
-	    CORBA_free(retptr);
+	    CORBA_free(*retptr);
+            g_free(retptr);
 	    break;
 	case CORBA_tk_struct:
 	case CORBA_tk_union:
 	    /* handle non-fixed size structs, arrays and unions like any/seq */
 	    if ((imethod->flags & ORBit_I_COMMON_FIXED_SIZE) == 0) {
-		CORBA_free(retptr);
+		CORBA_free(*retptr);
+                g_free(retptr);
 		break;
 	    }
 	    /* else fall through */
@@ -274,7 +283,7 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
     }
     if (argv) {
 	for (i = 0; i < num_args; i++) {
-	    gint flags = self->imethod->arguments._buffer[i].flags;
+	    gint flags = imethod->arguments._buffer[i].flags;
 
 	    if ((flags & ORBit_I_ARG_OUT) != 0)
 		CORBA_free(argvptr[i]);
@@ -284,7 +293,47 @@ pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
 	g_free(argv);
 	g_free(argvptr);
     }
-    g_free(arg_tc);
+}
+
+
+/* XXXX handle keyword arguments? */
+static PyObject *
+pycorba_method_call(PyCORBA_Method *self, PyObject *args, PyObject *kwargs)
+{
+    CORBA_Object objref;
+    CORBA_Environment ev;
+    CORBA_TypeCode ret_tc = TC_null;
+    gpointer ret = NULL, *retptr = NULL, *argv = NULL, *argvptr = NULL;
+    PyObject *obj, *pyret = NULL;
+    int num_args = 0, n_rets;
+
+    obj = PyTuple_GetItem(args, 0);
+    if (!PyObject_TypeCheck(obj, (PyTypeObject *)self->meth_class)) {
+	PyErr_SetString(PyExc_TypeError, "wrong object type as first arg");
+	return NULL;
+    }
+
+
+    if (!pycorba_call_marshal_args(self->imethod, args, &ret_tc, &ret,
+                                   &retptr, &argv, &argvptr, &num_args, &n_rets))
+        goto cleanup;
+
+
+    objref = ((PyCORBA_Object *)obj)->objref;
+
+    CORBA_exception_init(&ev);
+    pyorbit_begin_allow_threads;
+    ORBit_small_invoke_stub(objref, self->imethod, ret, argv,
+			    CORBA_OBJECT_NIL, &ev);
+    pyorbit_end_allow_threads;
+    if (pyorbit_check_ex(&ev))
+	goto cleanup;
+    CORBA_exception_free(&ev);
+
+    pyret = pycorba_call_demarshal_retval(self->imethod, num_args, ret_tc,
+                                          ret, argv, argvptr, n_rets);
+ cleanup:
+    pycorba_call_cleanup(self->imethod, num_args, ret, argv, argvptr, retptr, ret_tc);
     return pyret;
 }
 
@@ -457,6 +506,157 @@ pycorba_bound_method_call(PyCORBA_BoundMethod *self, PyObject *args, PyObject *k
     return ret;
 }
 
+typedef struct _PyORBitAsyncData {
+    PyObject *callback;
+    PyObject *user_data;
+    CORBA_TypeCode ret_tc;
+    gpointer ret;
+    gpointer *retptr;
+    gpointer *argv;
+    gpointer *argvptr;
+    int num_args;
+    int n_rets;
+} PyORBitAsyncData;
+
+static void
+async_callback (CORBA_Object          object,
+                ORBit_IMethod        *imethod,
+                ORBitAsyncQueueEntry *aqe,
+                gpointer              user_data,
+                CORBA_Environment    *ev)
+{
+    PyORBitAsyncData *async_data = user_data;
+    PyObject *py_async_retval = NULL, *pyexc_type = NULL, *pyexc_value = NULL;
+    PyObject *pytmp;
+    PyGILState_STATE state;
+
+    g_return_if_fail (async_data != NULL);
+
+    state = pyorbit_gil_state_ensure();
+    if (pyorbit_check_ex(ev)) {
+        PyObject *traceback = NULL;
+        PyErr_Fetch(&pyexc_type, &pyexc_value, &traceback);
+        Py_XDECREF(traceback);
+        goto call_python;
+    }
+
+    ORBit_small_demarshal_async (aqe, async_data->ret, async_data->argv, ev);
+    state = pyorbit_gil_state_ensure();
+    if (pyorbit_check_ex(ev)) {
+        PyObject *traceback = NULL;
+        PyErr_Fetch(&pyexc_type, &pyexc_value, &traceback);
+        Py_XDECREF(traceback);
+        goto call_python;
+    }
+
+    py_async_retval = pycorba_call_demarshal_retval(
+        imethod, async_data->num_args, async_data->ret_tc, async_data->ret,
+        async_data->argv, async_data->argvptr, async_data->n_rets);
+    pycorba_call_cleanup(imethod, async_data->num_args, async_data->ret,
+                         async_data->argv, async_data->argvptr,
+                         async_data->retptr, async_data->ret_tc);
+
+    if (pyorbit_check_ex(ev)) {
+        PyObject *traceback = NULL;
+        PyErr_Fetch(&pyexc_type, &pyexc_value, &traceback);
+        Py_XDECREF(traceback);
+        goto call_python;
+    }
+
+call_python:
+    if (!py_async_retval) {
+        Py_INCREF(Py_None);
+        py_async_retval = Py_None;
+    }
+    if (!pyexc_type) {
+        Py_INCREF(Py_None);
+        pyexc_type = Py_None;
+    }
+    if (!pyexc_value) {
+        Py_INCREF(Py_None);
+        pyexc_value = Py_None;
+    }
+    if (async_data->user_data)
+        pytmp = PyObject_CallFunction(async_data->callback, "NNNN", py_async_retval,
+                                      pyexc_type, pyexc_value, async_data->user_data);
+    else
+        pytmp = PyObject_CallFunction(async_data->callback, "NNN", py_async_retval,
+                                      pyexc_type, pyexc_value);
+    Py_DECREF(pytmp);
+    Py_DECREF(async_data->callback);
+    g_free(async_data);
+}
+
+
+static PyObject *
+pycorba_bound_method_async_call(PyCORBA_BoundMethod *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "args", "callback", "user_data", NULL};
+    PyObject *py_args, *py_callback, *py_user_data = NULL;
+    PyORBitAsyncData *async_data;
+    PyObject *selfarg, *tmp;
+    CORBA_Environment ev;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!O|O:async", kwlist,
+                                     &PyList_Type, &py_args, &py_callback, &py_user_data))
+        return NULL;
+
+    if (!PyCallable_Check(py_callback)) {
+        PyErr_SetString(PyExc_TypeError, "second argument not callable");
+        return NULL;
+    }
+
+    async_data = g_new0(PyORBitAsyncData, 1);
+    Py_INCREF(py_callback);
+    async_data->callback = py_callback;
+    Py_XINCREF(py_user_data);
+    async_data->user_data = py_user_data;
+
+    selfarg = PyList_New(1);
+    Py_INCREF(self->meth_self);
+    PyList_SET_ITEM(selfarg, 0, self->meth_self);
+    tmp = PySequence_Concat(selfarg, py_args);
+    if (!tmp)
+        PyErr_Print();
+    Py_DECREF(selfarg);
+    args = PySequence_Tuple(tmp);
+    Py_DECREF(tmp);
+
+    if (!pycorba_call_marshal_args(self->meth->imethod, args, &async_data->ret_tc, &async_data->ret,
+                                   &async_data->retptr, &async_data->argv, &async_data->argvptr,
+                                   &async_data->num_args, &async_data->n_rets))
+        goto error;
+
+    
+
+    CORBA_exception_init(&ev);
+
+    ORBit_small_invoke_async(((PyCORBA_Object *)self->meth_self)->objref,
+                             self->meth->imethod,
+                             async_callback, async_data,
+                             async_data->argv, CORBA_OBJECT_NIL, &ev);
+
+    if (ev._major != CORBA_NO_EXCEPTION) {
+        PyErr_SetString(PyExc_RuntimeError, "async invocation failed");
+        CORBA_exception_free(&ev);
+        goto error;
+    }
+
+    Py_INCREF(Py_None);
+    return Py_None;
+error:
+    Py_DECREF(args);
+    pycorba_call_cleanup(self->meth->imethod, async_data->num_args, async_data->ret,
+                         async_data->argv, async_data->argvptr, async_data->retptr,
+                         async_data->ret_tc);
+    return NULL;
+}
+
+static PyMethodDef pycorba_bound_method_methods[] = {
+    { "async", (PyCFunction)pycorba_bound_method_async_call, METH_KEYWORDS },
+    { NULL, NULL, 0 }
+};
+
 static PyObject *
 pycorba_bound_method_get_doc(PyCORBA_BoundMethod *self, void *closure)
 {
@@ -522,7 +722,7 @@ PyTypeObject PyCORBA_BoundMethod_Type = {
     0,                                  /* tp_weaklistoffset */
     (getiterfunc)0,                     /* tp_iter */
     (iternextfunc)0,                    /* tp_iternext */
-    0,                                  /* tp_methods */
+    pycorba_bound_method_methods,       /* tp_methods */
     0,                                  /* tp_members */
     pycorba_bound_method_getsets,       /* tp_getset */
     (PyTypeObject *)0,                  /* tp_base */
